--#############################################################
-- Author   : Boris Keil, Stefan Ritt
-- Contents : Status and control register definition for drs2_cmc2
-- $Id$
--#############################################################
-- changes: Wolfgang.Rauch@cern.ch 

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;

package dpnc342_pack is

-- Constant definitions
-- ####################

-- Wolfgang.Rauch@cern.ch
-- The DMA parameters must match the Cypress FX3 firmware

-- uc DMA IN  size 16384 = 0x4000 bytes => 4096 = 0x1000 32-bit words => 12-bit coding
  constant n_addr_bits_w32_usb_dma_in   : positive := 12;    
  constant n_w32_usb_dma_in             : positive := (2**n_addr_bits_w32_usb_dma_in);      

-- uc DMA OUT size 16384 = 0x4000 bytes => 4096 = 0x1000 32-bit words => 12-bit coding
  constant n_addr_bits_w32_usb_dma_out  : positive := 12;    
  constant n_w32_usb_dma_out            : positive := (2**n_addr_bits_w32_usb_dma_out);      

  constant n_addr_bits_w32_dpram_eeprom : positive := 14; -- number of address bits (32-bit words) -- 16k @ 32bit = 64k @ 8bit     
  constant n_addr_bits_w32_dpram        : positive := 12; -- number of address bits (32-bit words) --  4k @ 32bit = 16k @ 8bit     
  constant DEBUG_SIZE                   : positive := 6;  -- number of debug bits   -- TMP

  constant ECHO_BIT  : positive := 31;
  constant NOOP_BIT  : positive := 30;

  -- Register width. Do not modify.
  constant n_bits_byte     : integer := 8;
  constant n_bytes_reg     : integer := 4; 
  constant n_bits_reg      : integer := n_bytes_reg * n_bits_byte;

  constant n_addr_bits_w32_cmd_reg               : integer := 2;
  constant n_addr_bits_w32_ctrl_reg              : integer := 4;
  constant n_addr_bits_w32_stat_reg              : integer := 4;
--  constant const_pmc1_pmc2_diff_bit              : integer := 19;
  
  -- size of command array
  constant n_regs_cmd      : integer := (2**(n_addr_bits_w32_cmd_reg));
  -- number of control registers
  constant n_regs_ctrl     : integer := (2**(n_addr_bits_w32_ctrl_reg));
  -- number of status registers
  constant n_regs_stat     : integer := (2**(n_addr_bits_w32_stat_reg));
  -- reinit bit in control register #0
  constant c_reinit_bit    : integer := 1;

  constant n_drs_dacs     : positive := 5;
  constant n_bits_drs_dac : positive := 16;

  constant n_drs_adcs     : positive := 4;
  constant n_bits_drs_adc : positive := 12;

  -- "address space" layout
  constant c_ctrl_start        : unsigned(31 downto 0) := X"00000000";  
  constant c_ctrl_end          : unsigned(31 downto 0) := c_ctrl_start + (n_regs_ctrl * 4) - 1 ;  
  constant c_stat_start        : unsigned(31 downto 0) := X"00010000";  
  constant c_stat_end          : unsigned(31 downto 0) := c_stat_start + (n_regs_ctrl * 4) - 1 ; 
-- ram layout
-- eeprom         64k @ 8 bit: 0x00020000 - 0x0002ffff 
-- readout  ram 0 16k @ 8 bit: 0x00040000 - 0x00043fff 
-- readout  ram 1 16k @ 8 bit: 0x00044000 - 0x00047fff 
-- readout  ram 2 16k @ 8 bit: 0x00048000 - 0x0004bfff 
-- readout  ram 3 16k @ 8 bit: 0x0004c000 - 0x0004ffff 
-- pedestal ram 0 16k @ 8 bit: 0x00080000 - 0x00083fff 
-- pedestal ram 1 16k @ 8 bit: 0x00084000 - 0x00087fff 
-- pedestal ram 2 16k @ 8 bit: 0x00088000 - 0x0008bfff 
-- pedestal ram 3 16k @ 8 bit: 0x0008c000 - 0x0008ffff 
  constant c_ram_start         : unsigned(31 downto 0) := X"00020000";  
  constant c_ram_end           : unsigned(31 downto 0) := X"0008ffff";
  constant c_cmd_start         : unsigned(31 downto 0) := X"ffff0000";  
  constant c_cmd_end           : unsigned(31 downto 0) := c_cmd_start + (n_regs_cmd * 4) - 1 ;  
  
  -- SVN revision in hex format
  constant svn_revision : std_logic_vector(15 downto 0) := X"2b16";
  -- Board Serial Number in hex format
  constant board_sn : std_logic_vector(15 downto 0) := X"0002";

  -- size of DPRAM per PMC minus 4
--  constant const_dpram_size_m4: std_logic_vector(31 downto 0) := X"0000_9FFC"; -- 10 channels @ 1024 samples @ 2x12bits
  --constant const_dpram_size_m4: std_logic_vector(31 downto 0) := X"0000_0008"; -- for simulation

-- Type definitions
-- ################

  subtype t_small_delay is integer range 0 to 15;

  subtype t_reg is std_logic_vector(31 downto 0);

  type t_reg_cmd_a  is array (n_regs_cmd  - 1 downto 0) of t_reg;
  type t_reg_ctrl_a is array (n_regs_ctrl - 1 downto 0) of t_reg;
  type t_reg_stat_a is array (n_regs_stat - 1 downto 0) of t_reg;
  subtype t_cmd_reg_n  is integer range t_reg_cmd_a'low  to t_reg_cmd_a'high;
  subtype t_ctrl_reg_n is integer range t_reg_ctrl_a'low to t_reg_ctrl_a'high;
  subtype t_stat_reg_n is integer range t_reg_stat_a'low to t_reg_stat_a'high;

-- each control register has an associated "byte select" register
-- for 32-bit USB write accesses all n_bytes_per_reg bits should be set 
  type t_bytesel_a is array ((n_regs_ctrl - 1) downto 0) of std_logic_vector((n_bytes_reg-1) downto 0);

-- actions can be triggered by write accesses to certain bits in the command register array 
-- a write of a 1 to such a bit sets the corresponding bit in a bitset vector for ONE write cycle 
  type t_bitset_b is (bitset_start, bitset_reinit, bitset_soft, bitset_eeprom_w, bitset_eeprom_r);
  subtype t_bitset_v is std_logic_vector(t_bitset_b'pos(t_bitset_b'right) downto 0);

  type t_drs_dac_val_a is array (natural range <>) of std_logic_vector((n_bits_drs_dac-1) downto 0);
  type t_drs_adc_val_a is array (natural range <>) of std_logic_vector((n_bits_drs_adc-1) downto 0);

  type t_dpram_readout_addr_a is array (natural range <>) of std_logic_vector ((n_addr_bits_w32_dpram-1+2) downto 0);
  type t_dpram_readout_d_wr_a is array (natural range <>) of std_logic_vector( 31 downto 0);
  type t_dpram_readout_d_rd_a is array (natural range <>) of std_logic_vector( 31 downto 0);

--  constant type_trig_bit_len : natural := enum_t'pos(enum_t'right) + 1;

  type t_ar is (ar_ctrl, ar_stat, ar_ram, ar_cmd, ar_none); -- address ranges
  constant bs_reg_i     : natural := 0;  
  constant bs_eeprom_i  : natural := 13;  

end dpnc342_pack;

