----------------------------------------------------------------------------------
-- Company.......: Universit de Genve - DPNC
-- Author........: SD
-- File..........: spi_adc.vhd
-- Description...: SPI master to configure ADC, used to reset ADC after power-on
-- Project Name..: DPNC342 - 32-channel DRS4 Acquisition Board - NA61 New trigger
-- Created.......: Nov 3, 2014
-- Version.......: 0.1
--
-- Revision History:
-- Nov 3, 2014: file creation
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity spi_adc is
  port(CSB   : out   std_logic;
       SCLK  : out   std_logic;       -- max 25 MHz
       SDIO  : inout std_logic;
       SRST  : in    std_logic;       -- Soft reset
       CLK   : in    std_logic;
       RESET : in    std_logic
      );
end spi_adc;

--------------------------------------------------------------------------------
architecture bhv of spi_adc is

type state_type is (WAIT_PO,PHASE_H,PHASE_L,CMD_END,WAIT_NXT);
signal state : state_type;

subtype cnt_type is integer range 0 to 1048575;
constant PO_LATENCY   : cnt_type := 800000;    -- 20 ms @ CLK=40 MHz
constant WAIT_NXT_CMD : cnt_type := 200;

subtype i_type is integer range 0 to 7;
constant NUM_CMD : i_type := 5;

constant READ  : std_logic := '1';
constant WRITE : std_logic := '0';
constant WORD1 : std_logic_vector(1 downto 0) := "00";
constant WORD2 : std_logic_vector(1 downto 0) := "01";
constant WORD3 : std_logic_vector(1 downto 0) := "10";
constant WORD4 : std_logic_vector(1 downto 0) := "11";

constant PWRMOD_REG : std_logic_vector(7 downto 0) := X"08";
constant OUTMOD_REG : std_logic_vector(7 downto 0) := X"14";    -- to change output mode to offset binary
constant OUTPHA_REG : std_logic_vector(7 downto 0) := X"16";    -- output phase register
constant TSTMOD_REG : std_logic_vector(7 downto 0) := X"0d";    -- test mode register
constant RST_CMD    : std_logic_vector(7 downto 0) := X"03";    -- reset command
constant NOP_CMD    : std_logic_vector(7 downto 0) := X"00";    -- Normal Operation
constant OFB_CMD    : std_logic_vector(7 downto 0) := X"00";    -- to change output mode to offset binary
constant TCO_CMD    : std_logic_vector(7 downto 0) := X"01";    -- to change output mode to 2 complement (Default)
constant PHA_CMD    : std_logic_vector(7 downto 0) := X"03";    -- 180 degrees Phase adjustment (Default)
constant SYN_CMD    : std_logic_vector(7 downto 0) := X"0a";    -- 1x sync output test mode, output word 0x03f
constant TMO_CMD    : std_logic_vector(7 downto 0) := X"00";    -- output test mode off

type cmd_table is array (0 to NUM_CMD-1) of std_logic_vector(7 downto 0);

constant reg : cmd_table := (PWRMOD_REG,PWRMOD_REG,OUTMOD_REG,OUTPHA_REG,TSTMOD_REG);
constant cmd : cmd_table := (RST_CMD,NOP_CMD,OFB_CMD,PHA_CMD,TMO_CMD);

signal instruction : std_logic_vector(23 downto 0);

begin

spictl:process(CLK,RESET)
variable cnt : cnt_type;
variable i : i_type;
begin
  if (RESET='1') then
    CSB  <= '1';
    SCLK <= '0';
    SDIO <= 'Z';
    i := 0;
    cnt := 0;
    instruction <= (others => '0');
    state <= WAIT_PO;

  elsif (CLK'event AND CLK='1') then

    case state is

      when WAIT_PO =>
        if (cnt=PO_LATENCY) then
          CSB <= '0';
          instruction <= WRITE & WORD1 & "00000" & reg(i) & cmd(i);
          cnt := 0;
          state <= PHASE_H;
        else
          cnt := cnt + 1;
          state <= WAIT_PO;
        end if;

      when PHASE_H =>
        SCLK <= '0';
        SDIO <= instruction(23);
        instruction <= instruction(22 downto 0) & '0';
        state <= PHASE_L;

      when PHASE_L =>
        SCLK <= '1';
        if (cnt=23) then
          cnt := 0;
          state <= CMD_END;
        else
          cnt := cnt + 1;
          state <= PHASE_H;
        end if;

      when CMD_END =>
        SCLK <= '0';
        CSB  <= '1';
        SDIO <= 'Z';
        if (SRST='1') then
          i := 0;
          cnt := 0;
          state <= WAIT_PO;
        elsif (i < NUM_CMD-1) then
          i := i + 1;
          cnt := 0;
          state <= WAIT_NXT;
        else
          state <= CMD_END;
        end if;

      when WAIT_NXT =>
        if (cnt=WAIT_NXT_CMD) then
          CSB <= '0';
          instruction <= WRITE & WORD1 & "00000" & reg(i) & cmd(i);
          cnt := 0;
          state <= PHASE_H;
        else
          cnt := cnt + 1;
          state <= WAIT_NXT;
        end if;

    end case;
  end if;
end process spictl;

--------------------------------------------------------------------------------

end bhv;
