-- Inherited from S. Ritt et al. (usb2_racc.vhdl)
-- Modified by S. Debieux (usb2_racc.vhdl)
-- Incorporated code from Y. Favre
-- 20160912 further modified by Wolfgang.Rauch@cern.ch
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.dpnc342_pack.all;

-- ENTITY ----------------------------------------------------
entity usb3_racc is
  generic(		
    x_uc_data_bits         : positive  := 32;      -- uc interface datasize (must be 32 or 16)
    x_uc_dma_in_nb_w32     : positive;             -- uc DMA EP in buffer size (32-bit words)	
    x_uc_dma_out_nb_w32    : positive;             -- uc DMA EP out buufer size (32-bit words) 
    x_reset_active         : std_logic := '0'
    );
  port ( 
    x_clkin    : in std_logic;
    x_nreset   : in std_logic;
-- uC ports
    x_ucDQ     : inout std_logic_vector(x_uc_data_bits-1 downto 0);	-- uC FIFO slave data
    x_ucPCLK   : out   std_logic;       -- FX3 FIFO slave clock
    x_ucSLCSn  : out   std_logic;       -- FX3 FIFO slave chip select
    x_ucSLWRn  : out   std_logic;       -- FX3 FIFO slave write strobe
    x_ucSLRDn  : out   std_logic;       -- FX3 FIFO slave read strobe
    x_ucSLOEn  : out   std_logic;       -- FX3 FIFO slave output enable
    x_ucPKENDn : out   std_logic;       -- FX3 FIFO slave packet end
    x_ucADR    : out   std_logic_vector(1 downto 0);	-- FX3 FIFO slave address bus
    x_ucFLAG   : in    std_logic_vector(3 downto 0);	-- FX3 FIFO slave flags
    x_ucSTARTn : in    std_logic;       -- uC IO start signal	

-- Simple bus interface to on-chip RAM
-- --------------------------------------------------
    O_LOCBUS_ADDR  : out std_logic_vector(31 downto 0);
    I_LOCBUS_D_RD  : in  std_logic_vector(31 downto 0);
    O_LOCBUS_D_WR  : out std_logic_vector(31 downto 0);
    O_LOCBUS_WE    : out std_logic;     -- write enable for writing to dpram

-- Status & control registers
-- --------------------------
    O_CONTROL_REG_ARR       : out t_reg_ctrl_a;
    I_STATUS_REG_ARR        : in  t_reg_stat_a;

    O_BYTESEL_ARR           : out t_bytesel_a;
    O_BITSET_VEC            : out std_logic_vector(t_bitset_b'pos(t_bitset_b'right) downto 0);


-- information
    x_ready    : out std_logic		
--debug
--    x_debug    : out std_logic_vector((DEBUG_SIZE-1) downto 0)
    );
end usb3_racc;

-- ARCHITECTURE  ---------------------------------------------
architecture Behavioral of usb3_racc is

-- Constants -------------------------------------------------	
constant uc_cmd_bits       : positive := 8;
constant cmd_read          : std_logic_vector(uc_cmd_bits-1 downto 0)  := "00000001";
constant cmd_read_zlp      : std_logic_vector(uc_cmd_bits-1 downto 0)  := "00000101";
constant cmd_write         : std_logic_vector(uc_cmd_bits-1 downto 0)  := "00000010";

-- CAREFUL with delays, the constants are interpreted INCONSISTENTLY
--constant delay_start       : natural := 100000000; 
constant delay_start       : natural := 1000000; -- 10 ms 
constant delay_flaga_valid : t_small_delay := 4;   -- producer 
--constant delay_flaga_valid : natural := 10;    -- producer
constant delay_flagc_valid : t_small_delay := 3;     -- consumer 
--constant delay_flagc_valid : natural := 10;     -- consumer 
constant delay_cons_oe     : t_small_delay := 2;     -- consumer, read from uC
constant delay_read_ram    : t_small_delay := 1;     -- producer, read from FPGA ram

-- Types -----------------------------------------------------
type t_uc_state is (START, START1, START2, START3, START4, 
                       IDLE,
                       CONS, CONS_FLAGC_DELAY, CONS_FLAGC, CONS_OE_DELAY, CONS_01,
                       PROD, PROD_FLAGA_DELAY, PROD_FLAGA, PROD_RAM_DELAY, PROD_01,
                       DEAD);									
type t_uc_cmd_phase is (CMD_WORD,CMD_ADDRESS,CMD_BYTESIZE,CMD_VALID);
subtype t_pkt_in_size  is integer range 0 to (x_uc_dma_in_nb_w32-1);
subtype t_pkt_out_size is integer range 0 to (x_uc_dma_out_nb_w32-1);
subtype t_cmd_index    is integer range 0 to t_reg_cmd_a'high;


-- command related
constant iCMD_WORD       : t_cmd_index  := 0;
constant iCMD_ADDRESS    : t_cmd_index  := 1;
constant iCMD_BYTESIZE   : t_cmd_index  := 2;
constant iCMD_NUMRECV    : t_cmd_index  := 3;

-- Signals ---------------------------------------------------

signal x_reset : std_logic;

-- fake data
--signal data_s1         : std_logic_vector(31 downto 0);
--signal data_gen_all    : std_logic_vector(31 downto 0);
--signal data_gen_wren   : std_logic_vector(31 downto 0);
	
-- uC related
signal uc_state     : t_uc_state;
signal uc_cmd       : std_logic_vector(7 downto 0);
signal uc_data_i    : std_logic_vector(31 downto 0);
signal uc_data_o    : std_logic_vector(31 downto 0);
signal uc_data_o_ts : std_logic_vector(31 downto 0);
signal uc_pktend    : std_logic;
signal uc_flaga     : std_logic;
--signal uc_flagb     : std_logic;
signal uc_flagc     : std_logic;
--signal uc_flagd     : std_logic;
signal uc_start     : std_logic;
signal uc_fdts      : std_logic;
signal uc_slcs      : std_logic;
signal uc_fifoaddr  : std_logic_vector(1 downto 0);
signal uc_sloe      : std_logic;
signal uc_slrd      : std_logic;
signal uc_slwr      : std_logic;


-- internal control/status registers
signal delay_tmp             : natural;
signal cmd_reg_arr           : t_reg_cmd_a;
signal cmd_phase             : t_uc_cmd_phase;
signal addr_range            : t_ar;
signal uc_dma_in_cnt         : t_pkt_in_size;
signal uc_dma_out_cnt        : t_pkt_out_size;
signal control_reg_arr       : t_reg_ctrl_a;
signal status_reg_arr        : t_reg_stat_a;
signal usrbus_high           : signed(31 downto 0);
signal usrbus_i              : signed(31 downto 0);
signal locbus_addr           : std_logic_vector(31 downto 0);

-- components declaration ------------------------------------

component ibufreg
  generic (
    width : integer := 1
  );
  port (
    CLK  : in  std_logic;
    PRE  : in  std_logic;
    IPIN : in  std_logic_vector(width-1 downto 0);
    O    : out std_logic_vector(width-1 downto 0)
  );
end component;

component ibufreg1
  port (
    CLK  : in  std_logic;
    PRE  : in  std_logic;
    IPIN : in  std_logic;
    O    : out std_logic
  );
end component;

component ibufreg1r
  port (
    CLK  : in  std_logic;
    CLR  : in  std_logic;
    IPIN : in  std_logic;
    O    : out std_logic
  );
end component;

component obufreg
  generic (
    width : integer := 1
  );
  port (
    CLK  : in  std_logic;
    PRE  : in  std_logic;
    I    : in  std_logic_vector(width-1 downto 0);
    OPIN : out std_logic_vector(width-1 downto 0)
  );
end component;

component obufreg1
  port (
    CLK  : in  std_logic;
    PRE  : in  std_logic;
    I    : in  std_logic;
    OPIN : out std_logic
  );
end component;

-- function declaration ------------------------------------

BEGIN
-- Component Instantiation -----------------------------------	

  ibuf_data_i : ibufreg
  generic map (
    width => 32 
  )
  port map (
    CLK  => x_clkin,
    PRE  => x_reset, -- one presets to ones
    IPIN => x_ucDQ,
    O    => uc_data_i
  );

  ibuf_flaga : ibufreg1r
  port map (
    CLK  => x_clkin,
    CLR  => x_reset, -- one presets to 0
    IPIN => x_ucFLAG(0),
    O    => uc_flaga
  );

--  ibuf_flagb : ibufreg1r
--  port map (
--    CLK  => x_clkin,
--    CLR  => x_reset, -- one presets to 0
--    IPIN => x_ucFLAG(1),
--    O    => uc_flagb
--  );

  ibuf_flagc : ibufreg1r
  port map (
    CLK  => x_clkin,
    CLR  => x_reset, -- one presets to 0
    IPIN => x_ucFLAG(2),
    O    => uc_flagc
  );

-- ibuf_flagd : ibufreg1r
--  port map (
--    CLK  => x_clkin,
--    CLR  => x_reset, -- one presets to 0
--    IPIN => x_ucFLAG(3),
--    O    => uc_flagd
--  );

  ibuf_start : ibufreg1r
  port map (
    CLK  => x_clkin,
    CLR  => x_reset, -- one presets to 0
    IPIN => x_ucSTARTn,
    O    => uc_start
  );

-- combinational ---------------------------------------------

  x_ucPCLK   <= x_clkin; 
  x_reset    <= NOT x_nreset;

-- input
--  uc_data_i <= x_ucDQ;
--  uc_flaga  <= x_ucFLAG(0);
--  uc_flagb  <= x_ucFLAG(1);
--  uc_flagc  <= x_ucFLAG(2);
--  uc_flagd  <= x_ucFLAG(3);
--  uc_start  <= x_ucSTARTn;

-- output
  x_ucDQ       <= uc_data_o_ts when uc_fdts='0' else (others => 'Z');
  uc_data_o_ts <= uc_data_o;
  x_ucPKENDn   <= uc_pktend;
  x_ucSLOEn    <= uc_sloe;
  x_ucSLRDn    <= uc_slrd;
  x_ucSLWRn    <= uc_slwr;
  x_ucSLCSn    <= uc_slcs;
  x_ucADR      <= uc_fifoaddr;

  O_CONTROL_REG_ARR <= control_reg_arr;
  status_reg_arr    <= I_STATUS_REG_ARR;


-- process -----------------------------------------------
  p_clkin : process(x_nreset, x_clkin)
-- Variables 
  variable control_reg_no : t_ctrl_reg_n;

  begin
  if (x_nreset = x_reset_active) then

    cmd_reg_arr        <= (others => X"00000000");
    control_reg_arr    <= (others => X"00000000");
    locbus_addr        <= (others => '0');
    usrbus_high        <= (others => '0');
    usrbus_i           <= (others => '0');

    O_LOCBUS_ADDR      <= (others => '0');
    O_LOCBUS_D_WR      <= X"DEADBEEF";
    O_LOCBUS_WE        <= '0';

    O_BYTESEL_ARR      <= (others => "0000");
    O_BITSET_VEC       <= (others => '0');

    x_ready     <= '0';

-- debug

--    x_debug       <= (others => '0'); 

-- uC configuration

    uc_fdts     <= '1';  -- Tri-state data lines driven by uC
    uc_fifoaddr <= "11"; -- select uC consumer fifo (EP 01)

    uc_pktend   <= '1';
    uc_sloe     <= '1';
    uc_slrd     <= '1';
    uc_slwr     <= '1';
    uc_slcs     <= '1';
    
--    uc_flaga    <= '0';
--    uc_flagb    <= '0';
--    uc_flagc    <= '0';
--    uc_flagd    <= '0';

    uc_data_o   <= X"DEADBEEF";
--    uc_data_i   <= (others => '0');
    uc_state    <= START;

    uc_dma_in_cnt    <= 0;
    uc_dma_out_cnt   <= 0;

    delay_tmp <= 0;

  elsif (rising_edge(x_clkin)) then
--    data_gen_all      <= data_gen_all  + 1;
--    if(uc_slwr = '0') then 
--      data_gen_wren <= data_gen_wren + 1; 
--    end if;

      -- reset trigger signals
    O_BYTESEL_ARR      <= (others => "0000");
    O_BITSET_VEC          <= (others => '0');

      -- default values for RAM access
    O_LOCBUS_WE    <= '0';
    O_LOCBUS_D_WR  <= X"DEADBEEF";

      -- put ram address on bus on each clock cycle
    O_LOCBUS_ADDR <= locbus_addr;

      --   1. word:
      --     fd(7:0) : command
      --     "00000001"    cmd_read      => -- reading from FPGA, producing to uC EPin
      --     "00000101"    cmd_read_zlp  => -- reading from FPGA, producing to uC EPin
      --                                       generate ZLP if data ends at dma packet margin
      --     "00000010"    cmd_write     => -- consuming from uc EPout, writing to FPGA
      --   2. buffer:
      --     fd(31:0)  : address
      --   3. buffer:
      --     fd(31:0)  : byte size 
      --
      -- Address offsets:
      --   0x00000000     : CTRL
      --   0x00010000     : STATUS
      --   0x00020000     : FIFO
      --   0x00040000     : RAM BUF1
      --   0x00050000     : RAM BUF2
      --   0xffff0000     : CMD array
      -- ----------------
--    uc_data_i <= x_ucDQ;
    uc_pktend <= '1';
	
    case (uc_state) is 
      when START =>
        if(uc_start = '0') then uc_state <= START1; end if;
			
      when START1 => -- wait for 1st pulse (uncontrolled enum. reset or reboot from RAM)
        if(uc_start = '1') then uc_state <= START2; end if;

      when START2 =>
        if(uc_start = '0') then uc_state <= START3; end if;
			
      when START3 => -- wait for 2nd pulse (application set)
        if(uc_start = '1') then uc_state <= START4; end if;
			
      when START4 =>
        if(uc_start = '0') then 
		  uc_state <= IDLE; 
                  uc_slcs  <= '0';    -- chip select on
		  x_ready  <= '1';
		  delay_tmp <= delay_start;
		  --delay_tmp <= 0;
		  end if;

      when IDLE  =>
-- wait a little
-- there may be a problem with the flags not being asserted correctly 
-- right after the uC application has started
        if(delay_tmp > 0) then
          delay_tmp <= delay_tmp - 1;
	  uc_state  <= IDLE;
        else
          cmd_phase <= CMD_WORD;  -- initially we read a command
          uc_state  <= CONS;
        end if;

      when CONS  => 
-- when we enter the first time fifoaddr should have already been set to "11"
-- either thru the ibuf_flagc initalization or in a process
        uc_fifoaddr <= "11"; -- select uC consumer fifo
        uc_sloe     <= '1';  -- de-assert output enable
        uc_slrd     <= '1';  -- de-assert read enable
        uc_slwr     <= '1';  -- de-assert write enable
        uc_fdts     <= '1';  -- tri-state 
        uc_pktend   <= '1';  -- no packet end
--
-- AN65974
-- Designing with the EZ-USB ® FX3TM Slave FIFO Interface
-- When performing read transfers, a two-cycle latency for the flag is always incurred at the end of the transfer. 
-- The two- cycle latency is from the read (last SLRD# assertion) cycle that causes the buffer to become empty to the time 
-- the flag is asserted low. At the third clock edge, the external master can sample the flag low.
--
-- since we don't know where we're coming from we'll always wait...
-- we need two wait states
        delay_tmp <= delay_flagc_valid; -- do NOT lower this value
        uc_state  <= CONS_FLAGC_DELAY;  -- advance

      when CONS_FLAGC_DELAY => 
--        x_debug   <= "000001";
        if(delay_tmp > 0) then          -- consumer fifo flag may not be valid
            delay_tmp <= delay_tmp - 1; -- delay loop
        else
            uc_state  <= CONS_FLAGC;
        end if;

      when CONS_FLAGC  => 
        if(uc_flagc = '1') then       -- uC consumer fifo NOT empty
--          x_debug   <= "000011";
          uc_slrd   <= '0';     -- assert read enable
          uc_sloe   <= '0';     -- assert output enable for uC tri-state lines
          delay_tmp <= delay_cons_oe;
          uc_state  <= CONS_OE_DELAY;
        else
--          x_debug   <= "000010";
        end if;

      when CONS_OE_DELAY  => 
        if(delay_tmp > 0) then          -- delay between asserting uc_slrd and reading uc_data_i
          delay_tmp <= delay_tmp - 1;
        else
          uc_dma_out_cnt  <= t_pkt_out_size'HIGH ;  -- initialize counter for dma buffer out
          uc_state        <= CONS_01;
        end if;

-- we read from uc (consuming)
      when CONS_01  => 
--        x_debug           <= "000100";
-- we store our three command words in cmd_reg_arr
        case (cmd_phase) is
          when CMD_WORD     =>
	    cmd_reg_arr(iCMD_WORD)     <= uc_data_i;
            cmd_reg_arr(iCMD_NUMRECV)  <= cmd_reg_arr(iCMD_NUMRECV) + "1";  
            uc_cmd                     <= uc_data_i((uc_cmd_bits-1) downto 0);
--              cmd_read      => -- reading from FPGA, producing to uC EPin
--              cmd_read_zlp  => -- reading from FPGA, producing to uC EPin
--              cmd_write     => -- consuming from uc EPout, writing to FPGA
            cmd_phase     <= CMD_ADDRESS;  -- next we read ADDRESS

          when CMD_ADDRESS =>
	    cmd_reg_arr(iCMD_ADDRESS) <= uc_data_i;
            locbus_addr               <= uc_data_i; -- byte address
            if   ((unsigned(uc_data_i) >= c_ctrl_start) and 
                  (unsigned(uc_data_i) <= c_ctrl_end))     then addr_range <= ar_ctrl;
            elsif((unsigned(uc_data_i) >= c_stat_start) and 
                  (unsigned(uc_data_i) <= c_stat_end))     then addr_range <= ar_stat;
            elsif((unsigned(uc_data_i) >= c_ram_start)  and 
                  (unsigned(uc_data_i) <= c_ram_end))         then addr_range <= ar_ram;
            elsif((unsigned(uc_data_i) >= c_cmd_start)  and 
                  (unsigned(uc_data_i) <= c_cmd_end))         then addr_range <= ar_cmd;
            else
                addr_range <= ar_none;
              end if; 
              cmd_phase  <= CMD_BYTESIZE;  -- next we read byte size

          when CMD_BYTESIZE =>
	    cmd_reg_arr(iCMD_BYTESIZE) <= uc_data_i;
            -- from now on it is all command dependent
            case (uc_cmd)  is
              when cmd_read | cmd_read_zlp => -- reading from FPGA, producing to uC EPin
                cmd_phase    <= CMD_VALID;  -- command ready to be processed
             -- we are done with getting data from the uC
             -- we could not de-assert uc_slrd before so we do it now
             -- there are still two cycles after this one with uc_data_i driven by uC
             -- we don't care
                uc_slrd   <= '1';     -- de-assert read enable
                uc_sloe   <= '1';     -- de-assert output enable
-- usrbus_high may be -1 for nothing to send
-- even then we go to PROD because we want to send a ZLP
                usrbus_high  <= signed(("00" & uc_data_i(31 downto 2))-"1"); -- careful with uc_data_i = 0
                usrbus_i     <= (others => '0'); -- index for counting
                uc_state  <= PROD;
              when cmd_write => -- consuming from uc EPout, writing to FPGA
                if(uc_data_i(31 downto 2) = ("00" & X"0000000")) then
                -- nothing to write
                -- there are still two cycles after this one with uc_data_i driven by uC
                -- we don't care
                  cmd_phase    <= CMD_WORD;   -- command finished, go for next command
                  uc_slrd      <= '1';        -- de-assert read enable
                  uc_sloe      <= '1';        -- de-assert output enable
                  uc_state     <= CONS; 
                else
                  cmd_phase    <= CMD_VALID;  -- command ready to be processed
                  usrbus_high  <= signed(("00" & uc_data_i(31 downto 2))-"1"); -- careful with uc_data_i = 0
                  usrbus_i     <= (others => '0'); -- index for counting
                  uc_state     <= CONS_01; -- next lap we continue to read from uC and write to FPGA
                end if;
              when others => 
                cmd_phase <= CMD_WORD;       -- command finished
                uc_slrd   <= '1';        -- de-assert read enable
                uc_sloe   <= '1';        -- de-assert output enable
                uc_state  <= CONS;
            end case; -- (uc_cmd)

          when CMD_VALID =>
-- for now we must have got here because of a request to write to FPGA
            case (uc_cmd)  is
              when cmd_write => -- consuming from uc EPout, writing to FPGA
                case (addr_range) is
                  when ar_ctrl =>
-- for any change in the control register layout the following lines have to be modified accordingly
                    control_reg_no := conv_integer(locbus_addr(n_addr_bits_w32_ctrl_reg + 1 downto 2));
                    control_reg_arr(control_reg_no) <= uc_data_i;
                    O_BYTESEL_ARR(control_reg_no)((n_bytes_reg-1) downto 0) <= "1111";   -- one bit for each byte changed
                    if(control_reg_no=bs_reg_i) then
                       O_BITSET_VEC(t_bitset_b'pos(bitset_soft) downto t_bitset_b'pos(bitset_start))         <= uc_data_i(2 downto 0);
                    elsif(control_reg_no=bs_eeprom_i) then
                       O_BITSET_VEC(t_bitset_b'pos(bitset_eeprom_r) downto t_bitset_b'pos(bitset_eeprom_w))  <= uc_data_i(31 downto 30);
                    end if;
                    locbus_addr <= locbus_addr + "100"; -- 32-bit write
                  when ar_stat =>
                    locbus_addr <= locbus_addr + "100"; -- 32-bit write
                  when ar_ram =>
                    O_LOCBUS_WE    <= '1';
                    O_LOCBUS_D_WR  <= uc_data_i;
                    locbus_addr <= locbus_addr + "100"; -- 32-bit write
                  when ar_cmd =>
                    locbus_addr <= locbus_addr + "100"; -- 32-bit write
                  when others   =>
                    locbus_addr <= locbus_addr + "100"; -- 32-bit write
                end case; --  (addr_range)
-- the next if clause may not be that important
                if(uc_dma_out_cnt = delay_cons_oe) then
                  uc_slrd   <= '1';      -- de-assert read enable
                end if;
                if(uc_dma_out_cnt = 0) then
                  uc_state  <= CONS;
                  end if;                    
                if(usrbus_i = usrbus_high) then -- end of transfer
                  cmd_phase     <= CMD_WORD;       -- command finished
                  uc_slrd   <= '1';        -- de-assert read enable
                  uc_state  <= CONS;
                end if;
                  usrbus_i  <= usrbus_i + 1; -- update index

              when others => 
                cmd_phase     <= CMD_WORD;       -- command finished
                uc_slrd   <= '1';        -- de-assert read enable
                uc_sloe   <= '1';        -- de-assert output enable
                uc_state  <= CONS;
            end case; -- (uc_cmd)

          when others => 
            cmd_phase     <= CMD_WORD;       -- command finished
            uc_state  <= CONS;
          end case; -- (cmd_phase)
        uc_dma_out_cnt    <= uc_dma_out_cnt - 1;    -- update counter
-----
-----  END OF CONS_01
-----

-- FLAG A = 0: full,         uC producer fifo, thread 0, EP 1 IN
-- NOT USED -- FLAG B = 0: almost full,  uC producer fifo, thread 0, EP 1 IN   (watermark 7 ?) 
-- FLAG C = 0: empty,        uC consumer fifo, thread 3, EP 1 OUT
-- NOT USED -- FLAG D = 0: almost empty, uC consumer fifo, thread 3, EP 1 OUT  (watermark 6 ?)
-- GPIF burst parameter 1 in uC code according to YF. N.b. 2^1 = 2 
--
-- AN65974
-- Designing with the EZ-USB ® FX3TM Slave FIFO Interface
-- When performing write transfers, a three-cycle latency is always incurred at the end of the transfer. 
-- The three-cycle latency is from the write cycle that causes the buffer to become full to the time 
-- the flag is asserted low. At the fourth clock edge, the external master can sample the flag low. 
--
-- since we don't know where we're coming from we'll always wait...
-- we need at least three wait states, delay_tmp must have been set up.
-- 
      when PROD =>
-- to EP in 
--        x_debug       <= "001000";
        uc_data_o     <= X"DEADBEEF";
        uc_fifoaddr   <= "00"; -- select uC producer fifo, should have done beforehand
        uc_sloe       <= '1';  -- de-assert output enable
        uc_slrd       <= '1';  -- de-assert read enable
        uc_slwr       <= '1';  -- de-assert write enable
        uc_fdts       <= '1';  -- tri-state 
        delay_tmp     <= delay_flaga_valid; -- do NOT lower this value
        uc_state      <= PROD_FLAGA_DELAY;
     
      when PROD_FLAGA_DELAY => 
        if(delay_tmp > 0) then           -- producer fifo flag may not be valid
            delay_tmp <= delay_tmp - 1;  -- delay loop
        else
            uc_state  <= PROD_FLAGA;
        end if;

      when PROD_FLAGA => 
        uc_dma_in_cnt <= t_pkt_in_size'HIGH;
        if(uc_flaga = '1') then       -- uC producer fifo NOT full
-- beginning of a dma_in packet
-- zero size data or ZLP at dma_in packet end requested
          if(usrbus_i > usrbus_high) then -- end of transfer => ZLP
-- ZLP uc_slwr de-asserted, uc_pktend asserted
            uc_pktend     <= '0';      -- assert packet end
            uc_slwr       <= '1';      -- de-assert write enable
            cmd_phase     <= CMD_WORD; -- command finished 
            uc_state      <= CONS;     -- done
          else -- NOT a ZLP 
            if((addr_range = ar_ram) and (delay_read_ram > 0)) then -- NOT a ZLP but ram read with wait states
              delay_tmp <= delay_read_ram;
              uc_state  <= PROD_RAM_DELAY; 
            else -- NOT a ZLP, no wait states necessary
              uc_slwr       <= '0';     -- assert write enable
              uc_fdts       <= '0';     -- remove tri-state
              case (addr_range) is
                when ar_ctrl =>
                  uc_data_o   <= control_reg_arr(conv_integer(locbus_addr((n_addr_bits_w32_ctrl_reg + 1) downto 2)));
                  locbus_addr <= locbus_addr + "100"; -- 32-bit read
                when ar_stat =>
                  uc_data_o   <= status_reg_arr(conv_integer(locbus_addr((n_addr_bits_w32_stat_reg + 1) downto 2)));
                  locbus_addr <= locbus_addr + "100"; -- 32-bit read
                when ar_ram =>
                  uc_data_o   <= I_LOCBUS_D_RD;
                  locbus_addr <= locbus_addr + "100"; -- 32-bit read
                when ar_cmd =>
                  uc_data_o   <= cmd_reg_arr(conv_integer(locbus_addr((n_addr_bits_w32_cmd_reg + 1) downto 2)));
                  locbus_addr <= locbus_addr + "100"; -- 32-bit read
                when others   =>
                  uc_data_o   <= X"DEADBEEF";
              end case;
--              uc_data_o     <= (others => '0');
--              uc_data_o((x_uc_dma_in_size_w32-1) downto 0)     <= 
--                       std_logic_vector(to_unsigned(uc_dma_in_cnt,x_uc_dma_in_size_w32));
              if(usrbus_i = usrbus_high) then -- end of transfer
-- we cannot be at dma_in packet end, ZLP NOT necessary
                uc_pktend   <= '0';     -- assert packet end
                cmd_phase       <= CMD_WORD;    -- command finished 
                uc_state    <= CONS;    -- done
              else -- NOT end of transfer
                uc_dma_in_cnt <= uc_dma_in_cnt - 1; -- update dma buffer counter
                usrbus_i      <= usrbus_i      + 1; -- update data index
                uc_state      <= PROD_01; -- advance
              end if; -- NOT end of transfer
            end if; -- NOT a ZLP, no wait states necessary;
          end if; -- NOT a ZLP
--        else -- uC producer fifo IS full
--        uc_state       <= PROD_FLAGA;-- stay in this state
        end if;  -- uC producer fifo NOT full

      when PROD_RAM_DELAY =>
        locbus_addr <= locbus_addr + "100"; -- 32-bit read
        if(delay_tmp > 0) then
          delay_tmp   <= delay_tmp - 1;
        else
          uc_state    <= PROD_01;
        end if;

      when PROD_01 =>         
-- to EP in 
        uc_slwr         <= '0';     -- assert write enable
        uc_fdts         <= '0';     -- remove tri-state
        uc_data_o       <= X"DEADBEEF";
--        x_debug         <= "001100";
        case (addr_range) is
          when ar_ctrl =>
            uc_data_o   <= control_reg_arr(conv_integer(locbus_addr(n_addr_bits_w32_ctrl_reg + 1 downto 2)));
            locbus_addr <= locbus_addr + "100"; -- 32-bit read
          when ar_stat =>
            uc_data_o   <= status_reg_arr(conv_integer(locbus_addr(n_addr_bits_w32_stat_reg + 1 downto 2)));
            locbus_addr <= locbus_addr + "100"; -- 32-bit read
          when ar_ram =>
            uc_data_o <= I_LOCBUS_D_RD;
            if(uc_dma_in_cnt > delay_read_ram) then
-- Do NOT increment read address for the last few (delay_read_ram) ram reads
              locbus_addr <= locbus_addr + "100"; -- 32-bit read
            end if;
          when ar_cmd =>
            uc_data_o   <= cmd_reg_arr(conv_integer(locbus_addr(n_addr_bits_w32_cmd_reg + 1 downto 2)));
            locbus_addr <= locbus_addr + "100"; -- 32-bit read
          when others   =>
            uc_data_o   <= X"DEADBEEF";
        end case;
        if(uc_dma_in_cnt = 0) then  -- dma_in packet end
-- if on the host side we know the data size beforehand 
--        no ZLP will be required when the data ends at a dma packet margin
--        because we can issue a USB IN command with the exact data size
-- if on the host the data size is NOT known beforehand
--        we may want a ZLP when data ends on a dma packet margin
--        we use a special read command "cmd_read_zlp" for the request of a ZLP
          if((usrbus_i = usrbus_high) and (uc_cmd = cmd_read)) then -- dma_in packet end, end of transfer, not cmd_read_zlp
-------------- next two lines test
            uc_pktend   <= '0';      -- assert packet end
            cmd_phase   <= CMD_WORD;     -- command finished 
-------------- 
            uc_state <= CONS;
          else -- dma_in packet end, ZLP requested or more data
            uc_state <= PROD;   -- ZLP or more data
          end if;
        elsif(usrbus_i = usrbus_high) then    -- not dma_in packet end, end of transfer
-- NOT at dma_in packet end, ZLP NOT necessary
          uc_pktend   <= '0';      -- assert packet end
          cmd_phase   <= CMD_WORD;     -- command finished 
          uc_state    <= CONS;     -- done
        end if;
          uc_dma_in_cnt <= uc_dma_in_cnt - 1;       -- update counter
          usrbus_i      <= usrbus_i + 1;            -- update index
 
      when DEAD   =>
	x_ready  <= '0';
--	x_debug  <= "111111";
        uc_state <= DEAD;

      when others => 
        uc_state <= DEAD;
      end case;
    end if;
  end process;

-- FLAG A = 0: full,         uC producer fifo, thread 0, EP 1 IN
-- NOT USED -- FLAG B = 0: almost full,  uC producer fifo, thread 0, EP 1 IN   (watermark 7 ?) 
-- FLAG C = 0: empty,        uC consumer fifo, thread 3, EP 1 OUT
-- NOT USED -- FLAG D = 0: almost empty, uC consumer fifo, thread 3, EP 1 OUT  (watermark 6 ?)
	
end Behavioral;
